#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_oauth.zmi.plugin module

This modules is registering ZMI views which are used to configure OAuth authentication.
"""

from zope.interface import Interface

from pyams_auth_oauth.interfaces import IOAuthSecurityConfiguration
from pyams_auth_oauth.zmi.interfaces import IOauthConfigurationMenu
from pyams_form.ajax import ajax_form_config
from pyams_form.browser.checkbox import SingleCheckBoxFieldWidget
from pyams_form.field import Fields
from pyams_form.interfaces.form import IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_security_views.zmi import ISecurityMenu
from pyams_skin.interfaces.viewlet import IHeaderViewletManager
from pyams_skin.viewlet.help import AlertMessage
from pyams_utils.adapter import adapter_config
from pyams_viewlet.manager import viewletmanager_config
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminEditForm, FormGroupChecker
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem

__docformat__ = 'restructuredtext'

from pyams_auth_oauth import _  # pylint: disable=ungrouped-imports


@viewletmanager_config(name='oauth-security-configuration.menu',
                       context=ISecurityManager, layer=IAdminLayer,
                       manager=ISecurityMenu, weight=60,
                       provides=IOauthConfigurationMenu,
                       permission=MANAGE_SECURITY_PERMISSION)
class OAuthSecurityConfiguration(NavigationMenuItem):
    """OAUth security configuration menu"""

    label = _("OAuth configuration")
    href = '#oauth-security-configuration.html'


@ajax_form_config(name='oauth-security-configuration.html',
                  context=ISecurityManager, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class OAuthSecurityConfigurationEditForm(AdminEditForm):
    """OAUth security configuration edit form"""

    title = _("OAuth configuration")

    fields = Fields(Interface)


@adapter_config(name='oauth-configuration',
                required=(ISecurityManager, IAdminLayer, OAuthSecurityConfigurationEditForm),
                provides=IGroup)
class OAuthConfigurationGroup(FormGroupChecker):
    """OAuth configuration edit group"""

    fields = Fields(IOAuthSecurityConfiguration)
    fields['use_login_popup'].widget_factory = SingleCheckBoxFieldWidget

    def get_content(self):
        return IOAuthSecurityConfiguration(self.context)


@viewlet_config(name='oauth-configuration.header',
                context=ISecurityManager, layer=IAdminLayer, view=OAuthConfigurationGroup,
                manager=IHeaderViewletManager, weight=1)
class OauthConfigurationHeader(AlertMessage):
    """Oauth configuration header"""

    status = 'info'

    _message = _("""OAuth authentication module provides features and a REST API which can be \
used to authenticate principals using external OAuth providers.

Activating this option will require additional configuration to create a dedicated users folder \
to store profiles of users which will be registered using this protocol; you will also have to \
register API keys for providers that you want to allow connection from, before recording them \
in the security manager...
""")



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_oauth.zmi.folder module

This module holds views used to manage OAuth users folders.
"""

from pyramid.decorator import reify
from pyramid.view import view_config
from zope.interface import Interface

from pyams_auth_oauth.interfaces import IOAuthUser, IOAuthUsersFolderPlugin, \
    OAUTH_USERS_FOLDER_LABEL
from pyams_auth_oauth.plugin import get_provider_info
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces import ISecurityManager, IViewContextPermissionChecker
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_security_views.zmi import SecurityPluginsTable
from pyams_security_views.zmi.plugin import SecurityPluginAddForm, \
    SecurityPluginAddMenu, SecurityPluginPropertiesEditForm
from pyams_skin.interfaces.view import IModalPage
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextAdapter, ContextRequestViewAdapter, adapter_config
from pyams_utils.date import EXT_DATETIME_FORMAT
from pyams_utils.registry import get_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalDisplayForm
from pyams_zmi.helper.container import delete_container_element
from pyams_zmi.interfaces import IAdminLayer, TITLE_SPAN_BREAK
from pyams_zmi.interfaces.form import IFormTitle
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager
from pyams_zmi.search import SearchForm, SearchResultsView, SearchView
from pyams_zmi.table import DateColumn, I18nColumnMixin, IconColumn, Table, TableElementEditor, \
    TrashColumn
from pyams_zmi.utils import get_object_label

__docformat__ = 'restructuredtext'

from pyams_auth_oauth import _  # pylint: disable=ungrouped-imports


@viewlet_config(name='add-oauth-folder-plugin.menu',
                context=ISecurityManager, layer=IAdminLayer, view=SecurityPluginsTable,
                manager=IContextAddingsViewletManager, weight=60,
                permission=MANAGE_SECURITY_PERMISSION)
class OAuthFolderPluginAddMenu(SecurityPluginAddMenu):
    """OAuth folder plug-in add menu"""

    label = _("Add OAuth users folder...")
    href = 'add-oauth-folder-plugin.html'


@ajax_form_config(name='add-oauth-folder-plugin.html',
                  context=ISecurityManager, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class OAuthFolderPluginAddForm(SecurityPluginAddForm):
    """OAuth folder plug-in add form"""

    content_factory = IOAuthUsersFolderPlugin
    content_label = OAUTH_USERS_FOLDER_LABEL


@ajax_form_config(name='properties.html',
                  context=IOAuthUsersFolderPlugin, layer=IPyAMSLayer)
class OAuthFolderPropertiesEditForm(SecurityPluginPropertiesEditForm):
    """OAuth folder plug-in properties edit form"""

    plugin_interface = IOAuthUsersFolderPlugin


#
# OAuth users folder search view
#

class OAuthUsersSearchForm(SearchForm):  # pylint: disable=abstract-method
    """OAuth users search form"""

    title = _("Users search form")

    @property
    def back_url(self):
        """URL to previous page"""
        manager = get_utility(ISecurityManager)
        return absolute_url(manager, self.request, 'security-plugins.html')


@pagelet_config(name='search.html',
                context=IOAuthUsersFolderPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION)
class OAuthUsersSearchView(SearchView):
    """OAuth users search view"""

    title = _("Users search form")
    search_form = OAuthUsersSearchForm


class OAuthUsersSearchResultsTable(Table):
    """OAuth users search results table"""

    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        attributes['table'].update({
            'data-ams-order': '1,asc'
        })
        return attributes


@adapter_config(required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IValues)
class OAuthUsersSearchResultsValues(ContextRequestViewAdapter):
    """OAuth users search results values"""

    @property
    def values(self):
        """Search results values"""
        yield from self.context.get_search_results({
            'query': self.request.params.get('form.widgets.query')
        })


@adapter_config(name='provider',
                required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IColumn)
class OAuthUsersSearchProviderColumn(IconColumn):
    """OAuth users search provider column"""

    weight = 10

    def get_icon_class(self, item):
        provider_info = get_provider_info(item.provider_name)
        if provider_info is not None:
            return provider_info.icon_class
        return 'fa fa-question-circle'

    def get_icon_hint(self, item):
        provider_info = get_provider_info(item.provider_name)
        if provider_info is not None:
            return provider_info.name
        return None


@adapter_config(name='name',
                required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IColumn)
class OAuthUsersSearchNameColumn(I18nColumnMixin, GetAttrColumn):
    """OAuth users search name column"""

    i18n_header = _("Name")
    attr_name = 'name'
    default_value = '--'

    weight = 20


@adapter_config(name='email',
                required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IColumn)
class OAuthUsersSearchEmailColumn(I18nColumnMixin, GetAttrColumn):
    """OAuth users search email column"""

    i18n_header = _("Mail address")
    attr_name = 'email'
    default_value = '--'

    weight = 30


@adapter_config(name='registration_date',
                required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IColumn)
class OAuthUsersSearchRegistrationDateColumn(I18nColumnMixin, DateColumn):
    """OAuth users search registration date column"""

    i18n_header = _("Registration date")
    attr_name = 'registration_date'
    formatter = EXT_DATETIME_FORMAT

    weight = 50


@adapter_config(name='trash',
                required=(IOAuthUsersFolderPlugin, IAdminLayer, OAuthUsersSearchResultsTable),
                provides=IColumn)
class OAuthUsersSearchTrashColumn(TrashColumn):
    """OAuth users search trash column"""


@pagelet_config(name='search-results.html',
                context=IOAuthUsersFolderPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION, xhr=True)
class OAuthUsersSearchResultsView(SearchResultsView):
    """OAuth users search results view"""

    table_label = _("Search results")
    table_class = OAuthUsersSearchResultsTable


@view_config(name='delete-element.json',
             context=IOAuthUsersFolderPlugin, request_type=IPyAMSLayer,
             permission=MANAGE_SECURITY_PERMISSION, renderer='json', xhr=True)
def delete_local_user(request):
    """Local user delete view"""
    return delete_container_element(request)


#
# OAuth users views
#

@ajax_form_config(name='properties.html',
                  context=IOAuthUser, layer=IPyAMSLayer)
class OAuthUserEditForm(AdminModalDisplayForm):
    """OAuth user display form"""

    @property
    def subtitle(self):
        """Form title"""
        translate = self.request.localizer.translate
        return translate(_("User: {}")).format(self.context.title)

    legend = _("User properties")

    fields = Fields(IOAuthUser).omit('__parent__', '__name__')


@adapter_config(required=(IOAuthUser, IAdminLayer, IModalPage),
                provides=IFormTitle)
def oauth_user_form_title(context, request, form):
    """OAuth user form title"""
    translate = request.localizer.translate
    manager = get_utility(ISecurityManager)
    return TITLE_SPAN_BREAK.format(
        get_object_label(manager, request, form),
        translate(_("Plug-in: OAuth users folder")))


@adapter_config(required=(IOAuthUser, IAdminLayer, Interface),
                provides=ITableElementEditor)
class OAuthUserEditor(TableElementEditor):
    """OAuth users editor"""


@adapter_config(required=IOAuthUser, provides=IViewContextPermissionChecker)
class OAuthUserPermissionChecker(ContextAdapter):
    """OAuth user permission checker"""

    edit_permission = MANAGE_SECURITY_PERMISSION



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS OAuth authentication package interfaces module

"""

from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import containers, contains
from zope.interface import Attribute, Interface, Invalid, implementer, invariant
from zope.schema import Bool, Choice, Datetime, Int, List, Text, TextLine

from pyams_security.interfaces.plugin import IDirectorySearchPlugin


__docformat__ = 'restructuredtext'

from pyams_auth_oauth import _


OAUTH_LOGIN_ROUTE = 'pyams_oauth.login'
'''OAuth login route name'''

OAUTH_LOGIN_PATH = '/api/auth/oauth/{provider_name}'
'''OAuth login default path'''

OAUTH_CONFIGURATION_KEY = 'pyams_auth_oauth.configuration'
'''Main OAuth configuration key'''

OAUTH_LOGIN_CONFIGURATION_KEY = 'pyams_auth_oauth.providers'
'''OAuth providers configuration key'''

OAUTH_USERS_FOLDERS_VOCABULARY_NAME = 'pyams_auth_oauth.folders'
'''OAuth users folders vocabulary name'''

OAUTH_PROVIDERS_VOCABULARY_NAME = 'pyams_auth_oauth.providers'
'''OAuth providers vocabulary name'''


#
# OAuth's authentication utility interface
#

class IOAuthSecurityConfiguration(Interface):
    """Security manager configuration interface for OAuth"""

    enabled = Bool(title=_("Enable OAuth login?"),
                   description=_("Enable login via OAuth authentication"),
                   required=False,
                   default=False)

    users_folder = Choice(title=_("OAuth users folder"),
                          description=_("Name of folder used to store properties of users "
                                        "authenticated with OAuth"),
                          required=False,
                          vocabulary=OAUTH_USERS_FOLDERS_VOCABULARY_NAME)

    @invariant
    def check_users_folder(self):
        """Check for OAuth configuration"""
        if self.enabled and not self.users_folder:
            raise Invalid(_("You can't activate OAuth login without selecting an OAuth users "
                            "folder"))

    secret = TextLine(title=_("Authomatic secret"),
                      description=_("This secret phrase is used to encrypt Authomatic "
                                    "cookie"),
                      default='this is not a secret',
                      required=True)

    use_login_popup = Bool(title=_("Use OAuth popup?"),
                           description=_("If 'yes', a connection popup will be used"),
                           required=True,
                           default=False)


#
# OAuth login providers configuration
#

class IOAuthLoginProviderInfo(Interface):
    """OAuth login provider info

    This interface is used to adapt providers to
    get minimum information like icon class, URLs
    required to get consumer elements...
    """

    name = TextLine(title="Provider name")

    provider = Attribute("Provider class")

    icon_class = TextLine(title="Icon class",
                          description="Fontawesome icon class",
                          required=True)

    icon_filename = TextLine(title="Color icon filename",
                             required=True)

    scope = List(title="User info scope",
                 value_type=TextLine())


class IOAuthLoginConfiguration(Interface):
    """OAuth login configuration interface"""

    contains('pyams_auth_oauth.interfaces.IOAuthLoginProviderConnection')

    def get_oauth_configuration(self):
        """Get Authomatic configuration"""


class IOAuthLoginProviderConnection(Interface):
    """OAuth login provider info"""

    containers(IOAuthLoginConfiguration)

    provider_id = Int(title=_("Provider ID"),
                      description=_("This value should be unique between all providers"),
                      required=True,
                      readonly=True,
                      min=0)

    provider_name = Choice(title=_("Provider name"),
                           vocabulary=OAUTH_PROVIDERS_VOCABULARY_NAME,
                           required=True)

    consumer_key = TextLine(title=_("Provider consumer key"),
                            description=_("This consumer key is given by your OAuth provider..."),
                            required=True)

    consumer_secret = TextLine(title=_("Provider secret"),
                               description=_("This secret key is given by your OAuth "
                                             "provider..."),
                               required=True)

    access_headers = Text(title=_("Access headers"),
                          description=_("Some providers require custom headers; you can enter "
                                        "them in JSON format..."),
                          required=False)

    def get_configuration(self):
        """Get provider configuration"""


#
# OAuth users interfaces
#

OAUTH_USERS_FOLDER_LABEL = _("OAuth users folder plug-in")


class IOAuthUsersFolderPlugin(IDirectorySearchPlugin):
    """OAuth users folder interface"""

    contains('pyams_auth_oauth.interfaces.IOAuthUser')


class IOAuthUser(IAttributeAnnotatable):
    """OAuth user interface"""

    containers(IOAuthUsersFolderPlugin)

    user_id = TextLine(title=_("Internal provider ID"))

    provider_name = TextLine(title=_("OAuth provider name"))

    username = TextLine(title=_("User name"),
                        required=False)

    name = TextLine(title=_("Name"))

    first_name = TextLine(title=_('First name'),
                          required=False)

    last_name = TextLine(title=_('Last name'),
                         required=False)

    nickname = TextLine(title=_('Nickname'),
                        required=False)

    email = TextLine(title=_("E-mail address"),
                     required=False)

    timezone = TextLine(title=_('Timezone'),
                        required=False)

    country = TextLine(title=_('Country'),
                       required=False)

    city = TextLine(title=_('City'),
                    required=False)

    postal_code = TextLine(title=_("Postal code"),
                           required=False)

    locale = TextLine(title=_('Locale code'),
                      required=False)

    picture = TextLine(title=_('Picture URL'),
                       required=False)

    birth_date = Datetime(title=_('Birth date'),
                          required=False)

    registration_date = Datetime(title=_("Registration date"),
                                 readonly=True)


class IOAuthAuthenticationEvent(Interface):
    """OAuth's authentication event

    This event is notified after standard authentication, and can be used to redirect
    authenticated user to a registration form or any other location.
    """

    request = Attribute("Source request")

    plugin = Attribute("Event source plug-in")

    principal_id = Attribute("Authenticated principal ID")

    redirect_location = Attribute("Event redirect location")


@implementer(IOAuthAuthenticationEvent)
class OAuthAuthenticationEvent:
    """OAuth's authentication event"""

    def __init__(self, request, plugin, principal_id):
        self.request = request
        self.plugin = plugin
        self.principal_id = principal_id
        self.redirect_location = None



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_oauth.zmi.provider module

This module is registering views which are used to manage OAuth providers.
"""

import json
from json import JSONDecodeError

from pyramid.decorator import reify
from pyramid.events import subscriber
from pyramid.view import view_config
from zope.interface import Interface, Invalid

from pyams_auth_oauth.interfaces import IOAuthLoginConfiguration, IOAuthLoginProviderConnection
from pyams_auth_oauth.plugin import PROVIDERS_INFO
from pyams_auth_oauth.zmi.interfaces import IOauthConfigurationMenu
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer, IAddForm, IDataExtractedEvent
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces import ISecurityManager, IViewContextPermissionChecker
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_skin.interfaces.view import IModalPage
from pyams_skin.interfaces.viewlet import IHelpViewletManager
from pyams_skin.viewlet.actions import ContextAddAction
from pyams_skin.viewlet.help import AlertMessage
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextAdapter, ContextRequestViewAdapter, adapter_config
from pyams_utils.interfaces import MISSING_INFO
from pyams_utils.registry import get_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalAddForm, AdminModalEditForm
from pyams_zmi.helper.container import delete_container_element
from pyams_zmi.helper.event import get_json_table_row_add_callback, \
    get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer, TITLE_SPAN_BREAK
from pyams_zmi.interfaces.form import IFormTitle
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IToolbarViewletManager
from pyams_zmi.table import I18nColumnMixin, Table, TableAdminView, TableElementEditor, \
    TrashColumn
from pyams_zmi.utils import get_object_label
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem

__docformat__ = 'restructuredtext'

from pyams_auth_oauth import _  # pylint: disable=ungrouped-imports


@viewlet_config(name='oauth-providers.menu',
                context=ISecurityManager, layer=IAdminLayer,
                manager=IOauthConfigurationMenu, weight=10,
                permission=MANAGE_SECURITY_PERMISSION)
class OAuthProvidersMenu(NavigationMenuItem):
    """OAuth providers menu"""

    label = _("OAuth providers")
    href = '#oauth-providers.html'


class OAuthProvidersTable(Table):
    """OAuth providers table"""

    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        configuration = IOAuthLoginConfiguration(self.context)
        attributes['table'].update({
            'data-ams-location': absolute_url(configuration, self.request)
        })
        return attributes


@adapter_config(required=(ISecurityManager, IAdminLayer, OAuthProvidersTable),
                provides=IValues)
class OAuthProvidersTableValues(ContextRequestViewAdapter):
    """OAuth providers table values adapter"""

    @property
    def values(self):
        """Table values"""
        yield from IOAuthLoginConfiguration(self.context).values()


@adapter_config(name='id',
                required=(ISecurityManager, IAdminLayer, OAuthProvidersTable),
                provides=IColumn)
class OAuthProvidersIdColumn(I18nColumnMixin, GetAttrColumn):
    """OAuth providers ID column"""

    i18n_header = _("ID")
    attr_name = 'provider_id'

    weight = 10


@adapter_config(name='provider',
                required=(ISecurityManager, IAdminLayer, OAuthProvidersTable),
                provides=IColumn)
class OAuthProvidersProviderColumn(I18nColumnMixin, GetAttrColumn):
    """OAuth providers provider column"""

    i18n_header = _("Provider")
    attr_name = 'provider_name'

    weight = 20

    def get_value(self, obj):
        value = super().get_value(obj)
        try:
            return PROVIDERS_INFO[value].name
        except KeyError:
            return MISSING_INFO


@adapter_config(name='trash',
                required=(ISecurityManager, IAdminLayer, OAuthProvidersTable),
                provides=IColumn)
class OAuthProvidersTrashColumn(TrashColumn):
    """OAuth providers trash column"""


@pagelet_config(name='oauth-providers.html',
                context=ISecurityManager, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION, xhr=True)
class OAuthProvidersView(TableAdminView):
    """OAuth providers view"""

    table_label = _("Registered OAuth providers")
    table_class = OAuthProvidersTable


@viewlet_config(name='oauth-providers.help',
                context=ISecurityManager, layer=IAdminLayer, view=OAuthProvidersView,
                manager=IHelpViewletManager, weight=10)
class OAuthProvidersHelp(AlertMessage):
    """OAuth providers view help"""

    status = 'info'
    css_class = 'mx-2'
    _message = _("""You can provide as many OAuth connections as you want.
Each connection must have a unique integer ID.
""")


@view_config(name='delete-element.json',
             context=IOAuthLoginConfiguration, request_type=IPyAMSLayer,
             permission=MANAGE_SECURITY_PERMISSION, renderer='json', xhr=True)
def delete_oauth_provider(request):
    """OAuth provider delete view"""
    return delete_container_element(request)


@adapter_config(required=(IOAuthLoginProviderConnection, IAdminLayer, Interface),
                provides=ITableElementEditor)
class OAuthProviderElementEditor(TableElementEditor):
    """OAuth provider table element editor"""


@adapter_config(required=IOAuthLoginProviderConnection,
                provides=IViewContextPermissionChecker)
class OAuthProviderPermissionChecker(ContextAdapter):
    """OAuth provider permission checker"""

    edit_permission = MANAGE_SECURITY_PERMISSION


#
# OAuth provider views
#

@viewlet_config(name='add-oauth-provider.menu',
                context=ISecurityManager, layer=IAdminLayer, view=OAuthProvidersTable,
                manager=IToolbarViewletManager, weight=10,
                permission=MANAGE_SECURITY_PERMISSION)
class OAuthProviderAddAction(ContextAddAction):
    """OAuth provider add action"""

    label = _("Add provider")
    href = 'add-oauth-provider.html'

    def get_href(self):
        configuration = IOAuthLoginConfiguration(self.context)
        return absolute_url(configuration, self.request, self.href)


@ajax_form_config(name='add-oauth-provider.html',
                  context=IOAuthLoginConfiguration, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class OAuthProviderAddForm(AdminModalAddForm):
    """OAuth provider add form"""

    subtitle = _("New OAuth provider")
    legend = _("New OAuth provider properties")

    fields = Fields(IOAuthLoginProviderConnection).omit('__parent__', '__name__')
    content_factory = IOAuthLoginProviderConnection

    def update_content(self, obj, data):
        obj.provider_id = data.get(self, {}).get('provider_id')
        return super().update_content(obj, data)

    def add(self, obj):
        self.context[str(obj.provider_id)] = obj


@adapter_config(required=(IOAuthLoginConfiguration, IAdminLayer, IAddForm),
                provides=IFormTitle)
def oauth_provider_add_form_title(context, request, form):
    """OAuth provider add form title"""
    translate = request.localizer.translate
    manager = get_utility(ISecurityManager)
    return TITLE_SPAN_BREAK.format(
        get_object_label(manager, request, form),
        translate(_("Plug-in: OAuth authentication")))


@subscriber(IDataExtractedEvent, form_selector=OAuthProviderAddForm)
def extract_oauth_provider_add_form_data(event):
    """Check new provider data"""
    data = event.data
    configuration = event.form.context
    if str(data.get('provider_id')) in configuration:
        event.form.widgets.errors += (Invalid(_("This provider ID is already used!")),)


@adapter_config(required=(IOAuthLoginConfiguration, IAdminLayer, OAuthProviderAddForm),
                provides=IAJAXFormRenderer)
class OAuthProviderAddFormRenderer(ContextRequestViewAdapter):
    """OAUth provider add form AJAX renderer"""

    def render(self, changes):
        """JSON form renderer"""
        if not changes:
            return None
        sm = get_utility(ISecurityManager)
        return {
            'callbacks': [
                get_json_table_row_add_callback(sm, self.request,
                                                OAuthProvidersTable, changes)
            ]
        }


@ajax_form_config(name='properties.html',
                  context=IOAuthLoginProviderConnection, layer=IPyAMSLayer)
class OAuthProviderEditForm(AdminModalEditForm):
    """OAuth provider edit form"""

    @property
    def subtitle(self):
        try:
            provider = PROVIDERS_INFO[self.context.provider_name].name
        except KeyError:
            provider = MISSING_INFO
        translate = self.request.localizer.translate
        return translate(_("Provider: {}")).format(provider)

    legend = _("OAuth provider properties")

    fields = Fields(IOAuthLoginProviderConnection).omit('__parent__', '__name__')


@adapter_config(required=(IOAuthLoginProviderConnection, IAdminLayer, IModalPage),
                provides=IFormTitle)
def oauth_provider_edit_form_title(context, request, form):
    """Form title getter"""
    translate = request.localizer.translate
    manager = get_utility(ISecurityManager)
    return TITLE_SPAN_BREAK.format(
        get_object_label(manager, request, form),
        translate(_("Plug-in: OAuth authentication")))


@subscriber(IDataExtractedEvent, form_selector=OAuthProviderAddForm)
@subscriber(IDataExtractedEvent, form_selector=OAuthProviderEditForm)
def extract_oauth_provider_form_data(event):
    """Check new provider data"""
    data = event.data
    headers = data.get('access_headers')
    if headers:
        try:
            json.loads(headers)
        except JSONDecodeError:
            event.form.widgets.errors += (Invalid(_("Invalid JSON data for access headers!")),)


@adapter_config(required=(IOAuthLoginProviderConnection, IAdminLayer, OAuthProviderEditForm),
                provides=IAJAXFormRenderer)
class OAuthProviderEditFormRenderer(ContextRequestViewAdapter):
    """OAuth provider edit form renderer"""

    def render(self, changes):  # pylint: disable=missing-function-docstring
        if not changes:
            return None
        sm = get_utility(ISecurityManager)
        provider = self.view.context
        return {
            'callbacks': [
                get_json_table_row_refresh_callback(sm, self.request,
                                                    OAuthProvidersTable, provider)
            ]
        }



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_oauth.plugin module

This module is used to handle OAuth authentication.
"""

import json
from datetime import datetime

from authomatic.providers import oauth1, oauth2
from persistent import Persistent
from pyramid.events import subscriber
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary
from zope.traversing.interfaces import ITraversable

from pyams_auth_oauth.interfaces import IOAuthLoginConfiguration, IOAuthLoginProviderConnection, \
    IOAuthLoginProviderInfo, IOAuthSecurityConfiguration, IOAuthUser, IOAuthUsersFolderPlugin, \
    OAUTH_CONFIGURATION_KEY, OAUTH_LOGIN_CONFIGURATION_KEY, OAUTH_PROVIDERS_VOCABULARY_NAME, \
    OAUTH_USERS_FOLDERS_VOCABULARY_NAME
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.names import PRINCIPAL_ID_FORMATTER
from pyams_security.interfaces.plugin import IAuthenticatedPrincipalEvent
from pyams_security.interfaces.base import IPrincipalInfo
from pyams_security.principal import PrincipalInfo
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.request import check_request
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'


#
# OAuth security configuration
#

@factory_config(IOAuthSecurityConfiguration)
class OAuthSecurityConfiguration(Persistent, Contained):
    """OAuth security configuration"""

    enabled = FieldProperty(IOAuthSecurityConfiguration['enabled'])
    users_folder = FieldProperty(IOAuthSecurityConfiguration['users_folder'])
    secret = FieldProperty(IOAuthSecurityConfiguration['secret'])
    use_login_popup = FieldProperty(IOAuthSecurityConfiguration['use_login_popup'])


@adapter_config(required=ISecurityManager, provides=IOAuthSecurityConfiguration)
def security_manager_oauth_configuration_factory(context):
    """Security manager OAuth configuration factory adapter"""
    return get_annotation_adapter(context, OAUTH_CONFIGURATION_KEY, IOAuthSecurityConfiguration)


#
# OAuth user
#

@factory_config(IOAuthUser)
class OAuthUser(Persistent, Contained):
    # pylint: disable=too-many-instance-attributes
    """OAuth user persistent class"""

    user_id = FieldProperty(IOAuthUser['user_id'])
    provider_name = FieldProperty(IOAuthUser['provider_name'])
    username = FieldProperty(IOAuthUser['username'])
    name = FieldProperty(IOAuthUser['name'])
    first_name = FieldProperty(IOAuthUser['first_name'])
    last_name = FieldProperty(IOAuthUser['last_name'])
    nickname = FieldProperty(IOAuthUser['nickname'])
    email = FieldProperty(IOAuthUser['email'])
    timezone = FieldProperty(IOAuthUser['timezone'])
    country = FieldProperty(IOAuthUser['country'])
    city = FieldProperty(IOAuthUser['city'])
    postal_code = FieldProperty(IOAuthUser['postal_code'])
    locale = FieldProperty(IOAuthUser['locale'])
    picture = FieldProperty(IOAuthUser['picture'])
    birth_date = FieldProperty(IOAuthUser['birth_date'])
    registration_date = FieldProperty(IOAuthUser['registration_date'])

    @property
    def title(self):
        """Get user label"""
        if self.name:
            result = self.name
        elif self.first_name:
            result = '{} {}'.format(self.first_name, self.last_name or '')
        elif self.username:
            result = self.username
        else:
            result = self.nickname or self.user_id
        return result

    @property
    def title_with_source(self):
        """Get user label, including provider name"""
        return '{title} ({provider})'.format(title=self.title,
                                             provider=self.provider_name.capitalize())


@adapter_config(required=IOAuthUser, provides=IPrincipalInfo)
def oauth_user_principal_adapter(user):
    """OAuth user principal info adapter"""
    return PrincipalInfo(id="{}:{}".format(user.__parent__.prefix, user.user_id),
                         title=user.name)


#
# OAuth users folder
#

@factory_config(IOAuthUsersFolderPlugin)
class OAuthUsersFolder(Folder):
    """OAuth users folder"""

    prefix = FieldProperty(IOAuthUsersFolderPlugin['prefix'])
    title = FieldProperty(IOAuthUsersFolderPlugin['title'])
    enabled = FieldProperty(IOAuthUsersFolderPlugin['enabled'])

    def get_principal(self, principal_id, info=True):
        """Get principal with given ID"""
        if not self.enabled:
            return None
        if not (principal_id and principal_id.startswith(self.prefix + ':')):
            return None
        prefix, login = principal_id.split(':', 1)  # pylint: disable=unused-variable
        user = self.get(login)
        if user is not None:
            if info:
                return PrincipalInfo(id=PRINCIPAL_ID_FORMATTER.format(prefix=self.prefix,
                                                                      login=user.user_id),
                                     title=user.title)
            return user
        return None

    def get_all_principals(self, principal_id):
        """Get all principals for given principal ID"""
        if not self.enabled:
            return set()
        if self.get_principal(principal_id) is not None:
            return {principal_id}
        return set()

    def find_principals(self, query, exact_match=False):
        """Find principals matching given query"""
        if not self.enabled:
            return
        if not query:
            return
        query = query.lower()
        for user in self.values():
            for attr in (user.user_id, user.name, user.first_name,
                         user.last_name, user.email, user.nickname):
                if not attr:
                    continue
                if (exact_match and query == attr.lower()) or \
                        (not exact_match and query in attr.lower()):
                    yield PrincipalInfo(id='{0}:{1}'.format(self.prefix, user.user_id),
                                        title=user.title_with_source)
                    break

    def get_search_results(self, data):
        """Find principals matching given search query"""
        query = data.get('query')
        if not query:
            return
        query = query.lower()
        for user in self.values():
            if (query == user.user_id or
                    query in (user.name or '').lower() or
                    query in (user.email or '').lower()):
                yield user


@vocabulary_config(name=OAUTH_USERS_FOLDERS_VOCABULARY_NAME)
class OAuthUsersFolderVocabulary(SimpleVocabulary):
    """'PyAMS OAuth users folders' vocabulary"""

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        terms = []
        manager = query_utility(ISecurityManager)
        if manager is not None:
            for name, plugin in manager.items():
                if IOAuthUsersFolderPlugin.providedBy(plugin):
                    terms.append(SimpleTerm(name, title=plugin.title))
        super().__init__(terms)


@subscriber(IAuthenticatedPrincipalEvent, plugin_selector=IOAuthUsersFolderPlugin)
def handle_authenticated_oauth_principal(event):  # pylint: disable=invalid-name
    """Handle authenticated OAuth principal"""
    oauth_folder = event.plugin
    if (oauth_folder is None) or not oauth_folder.enabled:
        return
    infos = event.infos
    if not (infos and
            'provider_name' in infos and
            'user' in infos):
        return
    user = infos['user']
    principal_id = event.principal_id
    if ':' in principal_id:
        prefix, principal_id = principal_id.split(':', 1)
    if principal_id not in oauth_folder:
        oauth_user = OAuthUser()
        check_request().registry.notify(ObjectCreatedEvent(oauth_user))
        oauth_user.user_id = principal_id
        oauth_user.provider_name = infos['provider_name']
        oauth_user.username = user.username
        oauth_user.name = user.name
        oauth_user.first_name = user.first_name
        oauth_user.last_name = user.last_name
        oauth_user.nickname = user.nickname
        oauth_user.email = user.email
        oauth_user.timezone = str(user.timezone)
        oauth_user.country = user.country
        oauth_user.city = user.city
        oauth_user.postal_code = user.postal_code
        oauth_user.locale = user.locale
        oauth_user.picture = user.picture
        if isinstance(user.birth_date, datetime):
            oauth_user.birth_date = user.birth_date
        oauth_user.registration_date = datetime.utcnow()
        oauth_folder[principal_id] = oauth_user


#
# OAuth providers configuration
#

@implementer(IOAuthLoginProviderInfo)
class OAuthLoginProviderInfo:
    """OAuth login provider info"""

    name = FieldProperty(IOAuthLoginProviderInfo['name'])
    provider = None
    icon_class = FieldProperty(IOAuthLoginProviderInfo['icon_class'])
    icon_filename = FieldProperty(IOAuthLoginProviderInfo['icon_filename'])
    scope = FieldProperty(IOAuthLoginProviderInfo['scope'])

    def __init__(self, provider, **kwargs):
        self.provider = provider
        self.name = kwargs.get('name') or provider.__name__
        for key, value in kwargs.items():
            setattr(self, key, value)


PROVIDERS_INFO = {
    'amazon': OAuthLoginProviderInfo(provider=oauth2.Amazon,
                                     icon_class='fab fa-fw fa-amazon',
                                     icon_filename='amazon.ico',
                                     scope=oauth2.Amazon.user_info_scope),
    'behance': OAuthLoginProviderInfo(provider=oauth2.Behance,
                                      icon_class='fab fa-fw fa-behance-square',
                                      icon_filename='behance.ico',
                                      scope=oauth2.Behance.user_info_scope),
    'bitbucket': OAuthLoginProviderInfo(provider=oauth1.Bitbucket,
                                        icon_class='fab fa-fw fa-bitbucket',
                                        icon_filename='bitbucket.ico'),
    'bitly': OAuthLoginProviderInfo(provider=oauth2.Bitly,
                                    icon_class='fab fa-fw fa-bitly',
                                    icon_filename='bitly.ico',
                                    scope=oauth2.Bitly.user_info_scope),
    'cosm': OAuthLoginProviderInfo(provider=oauth2.Cosm,
                                   icon_class='fa fa-fw fa-share-alt',
                                   icon_filename='cosm.ico',
                                   scope=oauth2.Cosm.user_info_scope),
    'devianart': OAuthLoginProviderInfo(provider=oauth2.DeviantART,
                                        icon_class='fab fa-fw fa-deviantart',
                                        icon_filename='deviantart.ico',
                                        scope=oauth2.DeviantART.user_info_scope),
    'eventbrite': OAuthLoginProviderInfo(provider=oauth2.Eventbrite,
                                         icon_class='fa fa-fw fa-eventbrite',
                                         icon_filename='eventbrite.ico',
                                         scope=oauth2.Eventbrite.user_info_scope),
    'facebook': OAuthLoginProviderInfo(provider=oauth2.Facebook,
                                       icon_class='fab fa-fw fa-facebook-square',
                                       icon_filename='facebook.ico',
                                       scope=oauth2.Facebook.user_info_scope),
    'foursquare': OAuthLoginProviderInfo(provider=oauth2.Foursquare,
                                         icon_class='fab fa-fw fa-foursquare',
                                         icon_filename='foursquare.ico',
                                         scope=oauth2.Foursquare.user_info_scope),
    'flickr': OAuthLoginProviderInfo(provider=oauth1.Flickr,
                                     icon_class='fab fa-fw fa-flickr',
                                     icon_filename='flickr.ico'),
    'github': OAuthLoginProviderInfo(provider=oauth2.GitHub,
                                     icon_class='fab fa-fw fa-github',
                                     icon_filename='github.ico',
                                     scope=oauth2.GitHub.user_info_scope),
    'google': OAuthLoginProviderInfo(provider=oauth2.Google,
                                     icon_class='fab fa-fw fa-google',
                                     icon_filename='google.ico',
                                     scope=oauth2.Google.user_info_scope),
    'linkedin': OAuthLoginProviderInfo(provider=oauth2.LinkedIn,
                                       icon_class='fab fa-fw fa-linkedin-square',
                                       icon_filename='linkedin.ico',
                                       scope=oauth2.LinkedIn.user_info_scope),
    'meetup': OAuthLoginProviderInfo(provider=oauth1.Meetup,
                                     icon_class='fab fa-fw fa-meetup',
                                     icon_filename='meetup.ico'),
    'microsoft': OAuthLoginProviderInfo(provider=oauth2.MicrosoftOnline,
                                        icon_class='fab fa-fw fa-windows',
                                        icon_filename='microsoft.ico',
                                        scope=oauth2.MicrosoftOnline.user_info_scope),
    'paypal': OAuthLoginProviderInfo(provider=oauth2.PayPal,
                                     icon_class='fab fa-fw fa-paypal',
                                     icon_filename='paypal.ico',
                                     scope=oauth2.PayPal.user_info_scope),
    'plurk': OAuthLoginProviderInfo(provider=oauth1.Plurk,
                                    icon_class='fa fa-fw fa-share-alt',
                                    icon_filename='plurk.ico'),
    'reddit': OAuthLoginProviderInfo(provider=oauth2.Reddit,
                                     icon_class='fab fa-fw fa-reddit',
                                     icon_filename='reddit.ico',
                                     scope=oauth2.Reddit.user_info_scope),
    'tumblr': OAuthLoginProviderInfo(provider=oauth1.Tumblr,
                                     icon_class='fab fa-fw fa-tumblr-square',
                                     icon_filename='tumblr.ico'),
    'twitter': OAuthLoginProviderInfo(provider=oauth1.Twitter,
                                      icon_class='fab fa-fw fa-twitter',
                                      icon_filename='twitter.ico'),
    'ubuntuone': OAuthLoginProviderInfo(provider=oauth1.UbuntuOne,
                                        icon_class='fab fa-fw fa-ubuntu',
                                        icon_filename='ubuntuone.ico'),
    'viadeo': OAuthLoginProviderInfo(provider=oauth2.Viadeo,
                                     icon_class='fab fa-fw fa-viadeo',
                                     icon_filename='viadeo.ico',
                                     scope=oauth2.Viadeo.user_info_scope),
    'vimeo': OAuthLoginProviderInfo(provider=oauth1.Vimeo,
                                    icon_class='fab fa-fw fa-vimeo-square',
                                    icon_filename='vimeo.ico'),
    'vk': OAuthLoginProviderInfo(provider=oauth2.VK,
                                 icon_class='fab fa-fw fa-vk',
                                 icon_filename='vk.ico',
                                 scope=oauth2.VK.user_info_scope),
    'windowslive': OAuthLoginProviderInfo(provider=oauth2.WindowsLive,
                                          icon_class='fab fa-fw fa-windows',
                                          icon_filename='windows_live.ico',
                                          scope=oauth2.WindowsLive.user_info_scope),
    'xero': OAuthLoginProviderInfo(provider=oauth1.Xero,
                                   icon_class='fa fa-fw fa-share-alt',
                                   icon_filename='xero.ico'),
    'xing': OAuthLoginProviderInfo(provider=oauth1.Xing,
                                   icon_class='fab fa-fw fa-xing',
                                   icon_filename='xing.ico'),
    'yahoo': OAuthLoginProviderInfo(provider=oauth1.Yahoo,
                                    icon_class='fab fa-fw fa-yahoo',
                                    icon_filename='yahoo.ico'),
    'yammer': OAuthLoginProviderInfo(provider=oauth2.Yammer,
                                     icon_class='fab fa-fw fa-yammer',
                                     icon_filename='yammer.ico',
                                     scope=oauth2.Yammer.user_info_scope),
    'yandex': OAuthLoginProviderInfo(provider=oauth2.Yandex,
                                     icon_class='fab fa-fw fa-yandex',
                                     icon_filename='yandex.ico',
                                     scope=oauth2.Yandex.user_info_scope)
}


def get_provider_info(provider_name):
    """Get provider info matching given provider name"""
    return PROVIDERS_INFO.get(provider_name)


@vocabulary_config(name=OAUTH_PROVIDERS_VOCABULARY_NAME)
class OAuthProvidersVocabulary(SimpleVocabulary):
    """OAuth providers vocabulary"""

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        terms = []
        for key, provider in PROVIDERS_INFO.items():
            terms.append(SimpleTerm(key, title=provider.name))
        terms.sort(key=lambda x: x.title)
        super().__init__(terms)


@factory_config(IOAuthLoginConfiguration)
class OAuthLoginConfiguration(Folder):
    """OAuth login configuration"""

    def get_oauth_configuration(self):
        """Get OAuth configuration of registered providers"""
        result = {}
        for provider in self.values():
            provider_info = get_provider_info(provider.provider_name)
            provider_dict = {
                'id': provider.provider_id,
                'class_': provider_info.provider,
                'consumer_key': provider.consumer_key,
                'consumer_secret': provider.consumer_secret,
                'scope': provider_info.scope
            }
            if provider.access_headers:
                provider_dict['access_headers'] = json.loads(provider.access_headers)
            result[provider.provider_name] = provider_dict
        return result


@adapter_config(required=ISecurityManager, provides=IOAuthLoginConfiguration)
def oauth_login_configuration_adapter(context):  # pylint: disable=invalid-name
    """OAuth login configuration adapter"""
    return get_annotation_adapter(context, OAUTH_LOGIN_CONFIGURATION_KEY,
                                  IOAuthLoginConfiguration,
                                  name='++oauth-config++')


@adapter_config(name='oauth-config', required=ISecurityManager, provides=ITraversable)
class SecurityManagerOAuthTraverser(ContextAdapter):
    """++oauth-config++ namespace traverser"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Traverse to OAuth configuration"""
        return IOAuthLoginConfiguration(self.context)


@factory_config(IOAuthLoginProviderConnection)
class OAuthLoginProviderConnection(Persistent, Contained):
    """OAuth login provider connection"""

    provider_id = FieldProperty(IOAuthLoginProviderConnection['provider_id'])
    provider_name = FieldProperty(IOAuthLoginProviderConnection['provider_name'])
    consumer_key = FieldProperty(IOAuthLoginProviderConnection['consumer_key'])
    consumer_secret = FieldProperty(IOAuthLoginProviderConnection['consumer_secret'])
    access_headers = FieldProperty(IOAuthLoginProviderConnection['access_headers'])

    def get_configuration(self):
        """Get OAuth configuration of given provider"""
        return get_provider_info(self.provider_name)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_oauth.skin module

This module provides a login view for OAuth authentication.
Please note that this login method requires additional components provided by PyAMS_security_skin
package.
"""

from logging import WARNING

from authomatic import Authomatic
from authomatic.adapters import WebObAdapter
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound
from pyramid.response import Response
from pyramid.security import remember
from pyramid.view import view_config
from zope.interface import Interface

from pyams_auth_oauth.interfaces import IOAuthLoginConfiguration, IOAuthSecurityConfiguration, OAUTH_LOGIN_ROUTE, \
    OAuthAuthenticationEvent
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager, LOGIN_REFERER_KEY
from pyams_security.interfaces.names import PRINCIPAL_ID_FORMATTER
from pyams_security.interfaces.plugin import AuthenticatedPrincipalEvent
from pyams_security_views.interfaces.login import ILoginView
from pyams_skin.interfaces.viewlet import IFormFooterViewletManager
from pyams_template.template import template_config
from pyams_utils.registry import get_utility, query_utility
from pyams_viewlet.viewlet import Viewlet, viewlet_config


__docformat__ = 'restructuredtext'


@viewlet_config(name='oauth-providers.group',
                context=Interface, layer=IPyAMSLayer, view=ILoginView,
                manager=IFormFooterViewletManager, weight=10)
@template_config(template='templates/login-providers.pt', layer=IPyAMSLayer)
class OAuthProvidersGroup(Viewlet):
    """OAuth providers viewlet"""

    def __new__(cls, context, request, view, manager):  # pylint: disable=unused-argument
        sm = query_utility(ISecurityManager)  # pylint: disable=invalid-name
        if sm is None:
            return None
        configuration = IOAuthSecurityConfiguration(sm)
        if not configuration.enabled:
            return None
        return Viewlet.__new__(cls)

    @property
    def providers(self):
        """Get sorted OAuth providers"""
        sm = get_utility(ISecurityManager)  # pylint: disable=invalid-name
        configuration = IOAuthLoginConfiguration(sm)
        yield from sorted(configuration.values(),
                          key=lambda x: x.provider_id)


@view_config(route_name=OAUTH_LOGIN_ROUTE)
def login(request):
    """Login view for Authomatic authentication"""
    # check security manager utility
    manager = query_utility(ISecurityManager)
    if manager is None:
        raise HTTPNotFound()
    configuration = IOAuthSecurityConfiguration(manager)
    if not configuration.enabled:
        raise HTTPNotFound()
    # store referrer
    session = request.session
    if LOGIN_REFERER_KEY not in session:
        session[LOGIN_REFERER_KEY] = request.referer
    # init authomatic
    provider_name = request.matchdict.get('provider_name')
    # pylint: disable=assignment-from-no-return
    oauth_configuration = IOAuthLoginConfiguration(manager).get_oauth_configuration()
    authomatic = Authomatic(config=oauth_configuration,
                            secret=configuration.secret,
                            logging_level=WARNING)
    # perform login
    oauth_event = None
    response = Response()
    result = authomatic.login(WebObAdapter(request, response), provider_name)
    if result:
        if result.error:
            pass
        elif result.user:
            if not (result.user.id and result.user.name):
                result.user.update()
                if not (result.user.id and result.user.name):
                    raise HTTPBadRequest(result.user.content)
            oauth_folder = manager.get(configuration.users_folder)
            user_id = f'{provider_name}.{result.user.id}'
            principal_id = PRINCIPAL_ID_FORMATTER.format(prefix=oauth_folder.prefix,
                                                         login=user_id)
            request.registry.notify(AuthenticatedPrincipalEvent(plugin=oauth_folder,
                                                                principal_id=principal_id,
                                                                provider_name=provider_name,
                                                                user=result.user))
            headers = remember(request, principal_id)
            response.headerlist.extend(headers)
            oauth_event = OAuthAuthenticationEvent(request=request,
                                                   plugin=oauth_folder,
                                                   principal_id=principal_id)
            request.registry.notify(oauth_event)
        if configuration.use_login_popup:
            response.text = result.popup_html()
        response.status_code = 302
        if (oauth_event is not None) and oauth_event.redirect_location:
            response.location = oauth_event.redirect_location
        elif LOGIN_REFERER_KEY in session:
            response.location = session[LOGIN_REFERER_KEY]
            del session[LOGIN_REFERER_KEY]
        else:
            response.location = '/'
    return response
